"""
This module provides the version information for the package.

It attempts to import the version number generated by setuptools_scm.
If that fails, it defaults to "0.0.0".
"""

try:
    from ._version import version, __version__  # pylint: disable=unused-import
except ImportError:
    version = __version__ = "0.0.0"
